// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/material.dart';
import 'package:wyatt_ui_components/src/core/utils/theme_helper.dart';
import 'package:wyatt_ui_components/src/domain/entities/theme_style.dart';

/// {@template theme_resolver}
/// In charge of theme negotiation and merge.
///
/// When you build a component `Component({double? radius})`.
/// You have several possibilities:
/// 1) Pass the "radius" into the constructor, `Component(radius: 12)`.
/// 2) Set up a theme extension `ComponentThemeExtension(radius: 15)`.
/// 3) Let `wyatt_ui_kit` "negotiate" and try to find a suitable style in the
/// flutter theme, or use a hardcoded value.
///
/// If a negotiation phase fails, it will fallback to the next one.
///
/// This resolver uses [ThemeHelper] to negotiate and merge styles.
/// {@endtemplate}
abstract class ThemeResolver<Style extends ThemeStyle<Style>, Extension,
    Extra> {
  /// {@macro theme_resolver}
  const ThemeResolver();

  /// Compute default value from Flutter Theme or with hardcoded values.
  Style computeDefaultValue(
    BuildContext context, {
    Extra? extra,
  });

  /// Compute custom style from context.
  Style? Function(BuildContext context, {Extra? extra}) get customStyleFn;

  /// Compute extension value from custom component extension.
  Style? computeExtensionValueFn(
    BuildContext context,
    Extension? themeExtension, {
    Extra? extra,
  });

  /// Choose most suitable style for a given context.
  Style negotiate(BuildContext context, {Extra? extra}) {
    // 1) Custom style passed in constructor (cannot be null)
    final style = computeDefaultValue(context, extra: extra);
    return ThemeHelper.getElement<Style, Style>(
      [
        // 1) Custom style passed in constructor
        customStyleFn(context, extra: extra),
        // 2) Theme extension
        computeExtensionValueFn(
          context,
          Theme.of(context).extension<Extension>(),
          extra: extra,
        ),
        // 3) Default
        style,
      ],
      transform: (value) => value,
      combine: (value, element) => value?.mergeWith(element),
    );
  }
}
